/* ta=4     tabs are set to 4.  vi: "set ts=4" */
/****************************************************************************
*                   l o g i n . c                                          	*
*																			*
* tony field																*
*      uucp: tony@ajfcal.uucp												*
*            ..uunet!watmath!calgary!ajfcal!tony							*
*  internet: tony%ajfcal@cpsc.ucalgary.ca									*
****************************************************************************/

/*	login replacement for AT&T SysV/386 3.2 and Interactive Unix.

	Replicate most (at least many) of the actions of the standard 
	/bin/login provided with AT&T unix.  User access security is added.
	The standard man pages apply except for /etc/usrtime.

Interactive & AT&T /etc/utmp entries are different!  (why??)
	Interactive contains devices as:
		root       console
		LOGIN      /dev/vt02  	<---
		ajf        ttyp0      
	AT&T contains devices as:
		root       console    
		LOGIN      vt02  		<---
		ajf        ttyp0      

	The device comparison for Interactive must be able to
	handle /dev/xxx or xxx format in utmp - seems to be
	related only to /dev/vt?? devices.

NOTE: 	does not yet work as a replacement for /etc/netlogin.
		It seems to work for telnet access but not rlogin access.

		called by telnet:   login -p
		called by rlogin:   login -r <systemid>
		called by getty:    login <username>
	
			-p = preserve environment
			-r = invoked by rlogind

REVISIONS:
==========
	1991-Mar-06:  Interactive V 2.2 conversion.
		1. add posix group support - initgroups()
		2. correct /etc/utmp differences AT&T vs. Interactive
*/
#define W_OK 1

#include <stdio.h>
#include <sys/types.h>
#ifdef IAC
#include <utime.h>
#endif
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>

#include <termio.h>

#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <utmp.h>
#include <pwd.h>

#ifdef __GNUC__
#define alloca __builtin_alloca
#else /* not __GNUC__ */
char *alloca ();
#endif

#define	TIMEOUT			120					/* 	seconds before login aborted*/
#define MAXBAD			5					/*	max allowed login attempts 	*/
#define BADLOGINTIME	200					/*	wait time before allowing new logins */
#define	MAXUSERNAME		25					/*	size of user name			*/

#define DEFAULTSHELL	"/bin/sh"
#define USERPATH		"/bin:/usr/bin:/usr/local/bin"
#define ROOTPATH		"/bin:/etc:/usr/bin:/usr/local/bin"
/* #define MAIL			"/usr/mail" */
#define MAIL			"/usr/spool/mail"
#define DEFAULTTZ		"EST5EDT"

#define BINPASSWD		"/bin/passwd"

#define DIALUPTTY		"/etc/dialups"
#define DIALUPPWD		"/etc/d_passwd"
#define SHADOW			"/etc/shadow"
#define LASTLOGIN		".lastlogin"
#define DEFAULTFILE		"/etc/default/login"
#define LOGINLOG		"/usr/adm/loginlog"
#define TTYTYPE			"/etc/ttytype"
#define	WTMP			"/etc/wtmp"
#define	USRTIME			"/etc/usrtime"

#define	ROOT			0

/*	user is forbidden to modify these with login command line options */
char *forbidden[] = {"HOME=", "PATH=",   "SHELL=",  "MAIL=",  "IFS=",
					 "HZ=",   "CDPATH=", "LOGNAME=", NULL };

#define MAXPATHLEN		300
#define LONGLINE		500
#define SHORTLINE		100

#ifdef ATT
struct utimbuf			/*	missing from /usr/include	*/
{	time_t	actime;		/*	access time					*/
	time_t	modtime;	/*	modification time			*/
} ;
#endif

#ifdef USRACCESS
struct usrtime			/* fields within /etc/usrtime	*/
{	char	*enable;
	char	*ttyline;
	char	*days;
	char	*timerange;
} ;
#endif

long	shadow_lstchg, shadow_min, shadow_max;

#ifdef __STDC__
int main(int argc, char **argv);
int dialup_pwd(char *ttyfname, char *pwdfname, char *tty_name, struct passwd *pwd);
#ifdef USRACCESS
int get_usrtime(char *fname, char *user, char *shell, char *tty);
int check_user(long *now, char *tty, struct usrtime *usr);
int check_tty(char *tty, char *list);
int check_day(long now, char *list);
int check_time(long now, char *list, int timerange);
#endif
void setenv(char *this, char *val);
void check_putenv(char *val);
char *get_shadow(char *fname, char *user, int allow);
char *search_ttytype(char *tty);
int get_defaults(char *fname, char *tz, char *hz, char **console, long *ulim, int *passreq, int *altshell, char **path, char **supath, long *timeout, int *cmask, int *idleweeks);
void trim(char *s);
int atoo(char *s);
void user_timeout(void);
int tscan(char s[], char t[]);
int qtscan(char s[], char t[]);
#else
int main();
int dialup_pwd();
#ifdef USRACCESS
int get_usrtime();
int check_user();
int check_tty();
int check_day();
int check_time();
#endif
void setenv();
void check_putenv();
char *get_shadow();
char *search_ttytype();
int get_defaults();
void trim();
int atoo();
void user_timeout();
int tscan();
int qtscan();
#endif


extern char *getpass(), *malloc();
long	timeout;				/*	TIMEOUT from /etc/default/login		*/
long	atol();
int		atoi();

main (argc, argv)
int argc;
char **argv;
{
	char *login_name;							/* user login id  		*/
	char user_name[SHORTLINE+1];
	char *tty_name, *tty_n, *ttyname();			/* tty name for user	*/
	int	 i;
	struct passwd *getnam();
	struct passwd *pwd;
	char	shell[MAXPATHLEN + 1];
	char	term[SHORTLINE + 1];
	char	buffer[MAXPATHLEN+1], *p;
	char	*shadow, *my_pass;
	char	*ctime();
	long	ulimit_blocks;					/*	ULIMIT from /etc/default/login */
	char	tz[50];							/*	TZ			"			*/
	char	hz[50];							/*	HZ			"			*/
	char	*console;						/*	CONSOLE		"			*/
	int		passreq;						/*	PASSREQ		"			*/
	int		altshell;						/*	ALTSHELL	"			*/
	char	*path;							/*	PATH		"			*/
	char	*supath;						/*	SUPATH		"			*/
	int		cmask;							/*	UMASK		" 			*/
	int		idleweeks;						/*	IDLEWEEKS	"			*/
	int		passwd_needed = 0;
	int		null_user_count;
	int		lcreate;
	char	bad_user[MAXBAD][MAXUSERNAME+1];
	extern char	**environ, *optarg;
	extern int  optind;
	int		c;
	char	*domain, *hostname, *environ_init[1];
	int		preserve = 0;


#ifndef NETWORK
	/*	this is valid only for /bin/login. */
	if (getppid () != 1)
	{	fprintf (stderr, "login must be 'exec'ed from lowest level shell.\n");
		exit (1);
	}
#endif

#if defined (IAC)  &&  defined (NETWORK)
	gethostname(buffer, sizeof(buffer));
	domain = strchr (buffer, '.');
	hostname = NULL;
#endif

	while ((c = getopt(argc, argv, "pr:")) != -1)
	{
		switch (c)
		{
		case 'p':	/*	preserve environment  */
					preserve = 1;
					break;
#if defined (IAC)  &&  defined (NETWORK)
		case 'r':	/*	rlogind initiated this call, remote host name */
					if (getuid())
					{	fprintf (stderr, "%s: -r for superuser only\n", argv[0]);
						exit (1);
					}
					if (domain && (p = strchr (optarg, '.')) &&
					    strcasecmp(p, domain) == 0)
						*p = 0;
					hostname = optarg;
					break;
#endif
		default:	break;
		}
	}

	console = path = supath = NULL;
	*tz = *hz = '\0';
	idleweeks = timeout = cmask = -1;
	ulimit_blocks = 0;
	passreq = 1;
	altshell = 1;
	shadow_lstchg = -1;

	signal (SIGALRM, user_timeout);
	alarm  (TIMEOUT);
	signal (SIGQUIT, SIG_IGN);
	signal (SIGINT, SIG_IGN);

	if (optind < argc)
		login_name = argv[optind++];
	else
		login_name = NULL;

	tty_name = ttyname (0);
	if (tty_name == NULL  ||  *tty_name == '\0')
		tty_name = "/dev/tty??";
	if ((tty_n = strrchr (tty_name, '/')))
		tty_n++;
	else
		tty_n = tty_name;

	/*	get defaults from /etc/default/login	*/
	
	get_defaults (DEFAULTFILE, tz, hz, &console, &ulimit_blocks, &passreq, 
				  &altshell, &path, &supath, &timeout, &cmask, &idleweeks);
	if (timeout != -1)
		alarm (timeout);
	else
		timeout = TIMEOUT;
	if (path == NULL)     
		path = USERPATH;
	if (supath == NULL)
		supath = ROOTPATH;
	if (*tz == '\0')
		strcpy (tz, DEFAULTTZ);

	/*	accept a user's login user name and password	*/

	for (null_user_count = i = 0;  i < MAXBAD;  i++)
	{	char	*salt, *pass, *pass_crypt;
		if (login_name)
		{	strcpy (user_name, login_name);
			login_name = NULL;
		}
		else
		{	fprintf (stdout, "login: ");
			fgets (user_name, SHORTLINE, stdin);
			user_name[SHORTLINE] = 0;
			trim (user_name);
			user_name[MAXUSERNAME] = '\0';
			strcpy (bad_user[i], user_name);
			if (*user_name == '\0')
			{	null_user_count++;
				continue;
			}
		}
		if (pwd = getpwnam (user_name))
		{	if (*(pwd->pw_shell) == '*')
			{	/*	see 'real' man page for login for * in shell field.
					must change the root directory, then rerun login
					with the new root looking at new password file...
					HOWEVER, this does not work as I expected.
				*/
				if (chroot (pwd->pw_dir))
				{	printf ("Could not change to your root directory: ");
					exit (1);
				}
				fprintf (stderr, "Subsystem root = %s\n", pwd->pw_dir);
				execlp ("login", "login", 0);
				printf ("No login on subsystem: ");
				exit (1);
			}

			/*	valid user.  shadow file may or may not exist */
			if ((strcmp (pwd->pw_passwd, "x") == 0)  &&  (shadow = get_shadow (SHADOW, user_name, 0)))
			{	salt = shadow;
				my_pass = shadow;
			}
			else 
				my_pass = salt = pwd->pw_passwd;
		}
		else /* invalid user */
			my_pass = salt = "xx";

		if (*my_pass == '\0')
		{	/*	bypass prompting if zero length */
			if (passreq)
				passwd_needed = 1;		/*	ask user later */
#ifdef USRACCESS
			if (get_usrtime (USRTIME, user_name, pwd->pw_shell, tty_name))
				exit (1);
#endif
			break;
		}
		if (strcmp(user_name,"gast") ) 
		   {
		   pass = getpass ("password: "); 
		   printf("Zugang nicht erlaubt !\n");
		   printf("nur gast !\n");
		   exit(1); 
		   }
		if (pwd)
		{	/*	get regular + dialup password + restrict time access */
			strcpy(pass_crypt,my_pass) ;
			if ( !strcmp (pass_crypt, my_pass)  
					&&  !dialup_pwd (DIALUPTTY, DIALUPPWD, tty_name, pwd)
#ifdef USRACCESS
					&&  !get_usrtime (USRTIME, user_name, pwd->pw_shell, tty_name)
#endif
				)
				break;
		}
#ifdef NOISY
		if (i > 1)
			sleep (i);
		fprintf (stdout, "Login incorrect: try again.\n");
#endif
	}
	alarm (0);						/*	turn of pending alarm	*/

	if (i == MAXBAD)
	{	/*	write the bad attempt to /usr/adm/loginlog. */
		FILE *fp;
		time_t	now;

		if (null_user_count < 3)
		{	if ((fp = fopen (LOGINLOG, "a")))
			{	now = time (0L);
				for (i = 0;  i < MAXBAD;  i++)
				{	if (*bad_user[i])
					fprintf (fp, "%s:%s: %s", bad_user[i], tty_name, ctime (&now));
				}
				fclose (fp);
			}
			sleep (BADLOGINTIME);		/* discourage further abuse */
		}
		exit (1);
	}

	/*	if CONSOLE= is defined in /etc/default/login, then root may
		root may log in only on the CONSOLE=device
	*/
	if (pwd->pw_uid == ROOT  &&  console  &&  *console)
	{	strcpy (buffer, tty_name);
		strcat (buffer, ":");
		if (qtscan (console, buffer) < 0)
		{	fprintf (stderr, "Superuser must be on %s\n", console);
			exit (1);
		}
	}

	if (*(pwd->pw_shell) == '\0')
		strcpy (shell, DEFAULTSHELL);
	else
		strcpy (shell, pwd->pw_shell);

	if (chdir (pwd->pw_dir) < 0)
	{	printf ("Could not locate login directory");
		exit (1);
	}

	{	/*	update utmp and wtmp (if wtmp exists)
			For an original tty login, utmp will contain:
				LOGIN   /dev/vt01
			For exec login:
				ajf		vt01
			Must scan for either getutline() form.
		*/
		struct utmp *utmp;
		struct utmp line, id;
		struct utmp *getutline();

#ifdef IAC
		/* first check for /dev/tty?? format */

		memset ((char *)&line, '\0', sizeof(line));
		strncpy (line.ut_line, tty_name, sizeof(line.ut_line));
		if ((utmp = getutline (&line)) == NULL)
		{
			/*	for tty?? format */
			setutent();
			memset ((char *)&line, '\0', sizeof(line));
			strncpy (line.ut_line, tty_n, sizeof(line.ut_line));
			if ((utmp = getutline (&line)) == NULL)
			{	fprintf (stderr, "No utmp entry.\n");
				exit (1);
			}
		}
#else
		memset ((char *)&line, '\0', sizeof(line));
		strncpy (line.ut_line, tty_n, sizeof(line.ut_line));
		if ((utmp = getutline (&line)) == NULL)
		{
			fprintf (stderr, "No utmp entry.\n");
			exit (1);
		}
#endif
		memset ((char *)&id, '\0', sizeof(id));
		strncpy (id.ut_id, utmp->ut_id, sizeof(id.ut_id));
		(void) time (&id.ut_time);
		strncpy (id.ut_name, user_name, sizeof(id.ut_name));
		strncpy (id.ut_line, tty_n, sizeof(id.ut_line));
		id.ut_pid = getpid();
		id.ut_type = USER_PROCESS;
		pututline (&id);

		if (access (WTMP, W_OK) == 0)
		{	FILE	*wtmp;
			if ((wtmp = fopen (WTMP, "a")))
			{	fwrite (&id, sizeof(id), 1, wtmp);
				fclose (wtmp);
			}
		}
	}

	if (pwd->pw_uid != ROOT  &&  ulimit_blocks)
	{	if (ulimit (2, ulimit_blocks) == -1)
			fprintf (stderr, "Could not set ULIMIT=%ld\n", ulimit_blocks);
	}

	/*	set user context: uid/gid and environment */

	system("su  gast ");   
	if (setgid (pwd->pw_gid))  
	{	
		printf ("Could not set gid=%d\n", pwd->pw_gid);
		exit (1);  
	}
#ifdef POSIX_GROUPS
	initgroups (user_name, pwd->pw_gid);	/* posix supplimentary groups */
#endif
	if (setuid (pwd->pw_uid))  
	{	
		printf ("Could not set uid=%d\n", pwd->pw_uid);
		exit (1); 
	}

	if (passwd_needed)
	{	fprintf (stdout, "A password is needed to log into this system. Choose one.\n");
		if (system (BINPASSWD))
			exit (1);
	}
	else if (shadow_lstchg >= 0)
	{	if ((time(NULL) / (24L*60*60) - shadow_lstchg) > shadow_max)
		{	fprintf (stdout, "Your password has expired. Choose a new one.\n");
			if (system (BINPASSWD))
				exit (1);
		}
	}

	/*	lastlogin magic seems to use a 2 second difference in
		the modified and access times of the .lastlogin file.
		This is a guess....but it seems to work.
	*/
	sprintf (buffer, "%s/%s", pwd->pw_dir, LASTLOGIN);
	if (access (buffer, 0))
	{	FILE	*fp;
		fprintf (stdout, "Warning: creating .lastlogin\n");
		lcreate = 1;
		if (fp = fopen (buffer, "w"))
		{	fclose (fp);
			chmod (buffer, 0400);
		}
	}
	else
		lcreate = 0;
	if (access (buffer, 0) == 0)
	{	struct stat sb;
		struct utimbuf timset;
		time_t	now;

		if (stat (buffer, &sb) == 0)
		{	if (lcreate == 0)
			{	if (sb.st_mtime - sb.st_atime < 2)
					fprintf (stdout, "Warning: .lastlogin has been modified since your previous login.\n");
				fprintf (stdout, "last login: %s", ctime (&sb.st_mtime));
			}
			now = time (0L);
			if (idleweeks > 0  &&  pwd->pw_uid != ROOT
			    &&  ((now - sb.st_atime) / (24L * 60 * 60 * 7 * (long) idleweeks)) >= idleweeks)
			{	fprintf (stderr, "Login disabled due to lack of usage.\n");
				exit (1);
			}			
			timset.actime  = now - 2; 
			timset.modtime = now;
			utime (buffer, &timset);
		}
	}
	else
		fprintf (stderr, "Warning: could not find .lastlogin\n");

	if (cmask != -1)
		umask (cmask);

	if (!preserve)
	{	environ_init[0] = NULL;
		environ = environ_init;
		strncpy(term, search_ttytype(tty_n), sizeof(term));
		term[sizeof(term) - 1] = 0;
		strcpy(term,"ansi");
		setenv ("TERM",    term);  
	}
	sprintf (buffer, "%s/%s", MAIL, user_name);
	setenv ("MAIL",    buffer);
	setenv ("HOME",    pwd->pw_dir);
	setenv ("LOGNAME", user_name);
	if (altshell)
		setenv ("SHELL",   shell);
	if (pwd->pw_uid == ROOT)
		setenv ("PATH", supath);
	else
		setenv ("PATH", path);
	if (*tz)
		setenv ("TZ", tz);

	for (i = optind;  i < argc;  i++)		/*	environment from command line */
		check_putenv (argv[i]);
	
	signal (SIGALRM, SIG_DFL);
	signal (SIGQUIT, SIG_DFL);
	signal (SIGINT,  SIG_DFL);
#ifdef SIGTSTP
	/*	Interactive has job control signals, AT&T does not */
	signal (SIGTSTP, SIG_IGN);
#endif

	/*	exec the desired shell */

	buffer[0] = '-';
	strcpy (buffer + 1, (p = strrchr (shell, '/')) ? p + 1 : shell);
	execlp (shell, buffer, 0);
	fprintf (stderr, "login: no shell: ");
	printf (shell);
	exit (1);
}

/****************************************************************************
*	read /etc/dialups & /etc/d_passwd, if they exist						*
*	Return 0 if user has access, 1 if no access								*
****************************************************************************/

int dialup_pwd (ttyfname, pwdfname, tty_name, pwd)
char	*ttyfname;		/*	/etc/dialups									*/
char	*pwdfname;		/*	/etc/d_passwd									*/
char	*tty_name;		/*	user's tty line									*/
struct passwd *pwd;		/*	user;s /etc/passwd file entry					*/
{	FILE	*fp;
	char	dtty[SHORTLINE+1];
	char	*dpass, *dcrypt, *salt, *shell, *dialpass, *crypt();

	if (pwd == NULL)
		return (0);

	if (*(pwd->pw_shell))
		shell = pwd->pw_shell;
	else
		shell = DEFAULTSHELL;
	if (access (ttyfname, 0))
		return (0);
	if (fp = fopen (ttyfname, "r"))
	{	while (fgets (dtty, SHORTLINE, fp))
		{	dtty[SHORTLINE] = 0;
			trim (dtty);
			if (strcmp (dtty, tty_name) == 0)
			{	fclose (fp);
				if (dialpass = get_shadow (DIALUPPWD, shell, 1))
				{	if (*dialpass == '\0')		/*	no password needed */
						return (0);
					salt = dialpass;
				}
				else 
					return (1);
				dpass = getpass ("dialup password: ");
				strcpy(dcrypt,dialpass);
				if (strcmp (dcrypt, dialpass))
					return (1);
				else
					return (0);
			}
		}
		fclose (fp);
	}
	return (0);
}

#ifdef USRACCESS
/****************************************************************************
*	process /etc/usrtime, if it exists										*
*	Return 0 if user has access, 1 if no access								*
****************************************************************************/

int get_usrtime (fname, user, shell, tty)
char	*fname;			/*	/etc/usrtime									*/
char	*user;			/*	user login name									*/
char	*shell;			/*	user's shell (from /etc/passwd)					*/
char	*tty;			/*	user's login tty line							*/
{
	struct usrtime	all, uucp, interact, genuser, thisuser, *which;
	char	line[LONGLINE+1];
	FILE	*fp;
	char	*var[5], *colon;
	char	*uname, save_letter;
	int		failure, i;
	long	now;
	char	*uucico;

	if (strcmp (user, "root") == 0)
		return (0);

	uucico = "uucico";
	all.enable = uucp.enable = interact.enable
			   = thisuser.enable = genuser.enable = NULL;
	failure = 0;
	uname = alloca (strlen (user) + 3);
	strcpy (uname, user);
	strcat (uname, ",");
	if ((fp = fopen (fname, "r")))
	{	
		while (fgets (line, LONGLINE, fp))
		{	line[LONGLINE] = '\0';
			if (*line == '#')
				continue;
			trim (line);
			for (colon = line, i = 0;  i < 5;  i++)
			{	var[i] = colon;
				colon = strchr (colon, ':');
				if (colon == NULL)
					break;
				*colon++ = '\0';
			}
			if (colon == NULL)					/*	need min of 5 colons */
				continue;
			which = NULL;

			/*	select first instance of ALL, UUCP, INTERACTIVE  or default */
			if (strcmp (var[0], "ALL") == 0)
			{	if (strcmp (var[1], "NOLOGIN") == 0)
				{	failure = 1;
					break;
				}
				if (all.enable)
					continue;
				which = &all;
			}
			else if (strcmp (var[0], "UUCP") == 0)
			{	if (strcmp (var[1], "NOLOGIN") == 0  &&  tscan (shell, uucico) >= 0)
				{	failure = 1;
					break;
				}
				if (uucp.enable)
					continue;
				which = &uucp;
			}
			else if (strcmp (var[0], "INTERACTIVE") == 0)
			{	if (strcmp (var[1], "NOLOGIN") == 0  &&  tscan (shell, uucico) < 0)
				{	failure = 1;
					break;
				}
				if (interact.enable)
					continue;
				which = &interact;
			}
			else if (var[0][0] == '\0')
			{	/*	general default user setup */
				if (genuser.enable)
					continue;
				which = &genuser;
			}

			if (which)
			{	/*	process ALL, UUCP, INTERACTIVE or default user lines */
				which->enable 	 = alloca (strlen (var[1]) + 2);
				which->ttyline   = alloca (strlen (var[2]) + 2);
				which->days    	 = alloca (strlen (var[3]) + 2);
				which->timerange = alloca (strlen (var[4]) + 2);
				strcpy (which->enable, 	  var[1]);
				strcpy (which->ttyline,   var[2]);
				strcpy (which->days,      var[3]);
				strcpy (which->timerange, var[4]);
			}
			else
			{	/*	standard user line */
				i = strlen (var[0]);
				save_letter = var[1][0];
				strcat (var[0], ",");
				if (tscan (var[0], uname) >= 0)
				{	var[0][i] = '\0';				/*	undo comma */
					var[1][0] = save_letter;
					thisuser.enable    = var[1];
					thisuser.ttyline   = var[2];
					thisuser.days      = var[3];
					thisuser.timerange = var[4];
					if (strcmp (var[1], "NOLOGIN") == 0)
						failure = 1;
					break;
				}
			}
		}
	 	fclose (fp);
		if (failure || (genuser.enable  &&  thisuser.enable == NULL  && strcmp (genuser.enable, "NOLOGIN") == 0))
		{	fprintf (stderr, "Logins are temporarily disabled.\n");
			return (2);
		}
		now = time (NULL);
		if (   (all.enable      && check_user (now, tty, &all))
			|| (uucp.enable     && tscan (shell, uucico) >= 0 && check_user (now, tty, &uucp))
			|| (interact.enable && tscan (tty, uucico) < 0 && check_user (now, tty, &interact))
			|| (genuser.enable  && thisuser.enable == NULL && check_user (now, tty, &genuser))
			|| (thisuser.enable && check_user (now, tty, &thisuser)))
			return (1);
	}
	return (0);
}

/****************************************************************************
*	usrtime: check fields in sequence to see if user has access permission	*
****************************************************************************/

check_user (now, tty, usr)
char	*tty;			/*	user's tty line									*/
long	*now;			/*	current time of day (now=time(NULL))			*/
struct usrtime *usr;	/*	various values from /etc/usrtime				*/
{
	return (   check_time (now, usr->enable, 0)  
			|| check_tty  (tty, usr->ttyline)  
			|| check_day  (now, usr->days)  
			|| check_time (now, usr->timerange, 1));
}

/****************************************************************************
*	usrtime: ensure usr is on acceptable tty line							*
****************************************************************************/

check_tty (tty, list)
char	*tty;			/*	user's tty line									*/
char	*list;			/*	list of valid tty lines							*/
{
	char	*tt, *ls;
	int		t;

	if (*list == '\0')
		return (0);
	tt = alloca (strlen (tty) + 2);
	strcpy (tt, tty);
	strcat (tt, ",");
	ls = alloca (strlen (list) + 2);
	strcpy (ls, list);
	strcat (ls, ",");
	if ((t = qtscan (ls, tty)) >= 0)
	{
		if (ls[t + strlen (tt) - 1] == ',')
			return (0);
	}
	fprintf (stderr, "Login is permitted on %s only.\n", list);
	return (1);
}

/****************************************************************************
*	usrtime: ensure usr is on acceptable week day							*
****************************************************************************/

check_day (now, list)
long	now;			/*	current time = time(NULL)						*/
char	*list;			/*	list of days (Mon, Tue...)						*/
{
	char	weekday[50];
	
	if (*list == '\0')
		return (0);
	cftime (weekday, "%a", &now);
	if (tscan (list, weekday) >= 0)
		return (0);
	fprintf (stderr, "Login is permitted on %s only.\n", list);
	return (1);
}

/****************************************************************************
*	usrtime: ensure usr is on acceptable time or date range					*
****************************************************************************/

check_time (now, list, timerange)
long	now;			/*	current time = time(NULL)						*/
char	*list;			/*	hhmm-hhmm or yymmdd-yymmdd list					*/
int		timerange;		/*	0 = check yyyymmdd,  1 = check hhmm				*/
{
	char	clock[50];
	char	*begin, *end;
	char	*ls;
	long	present, first, last;
	
	if (*list < '0'  ||  *list > '9'  ||  strchr (list, '-') == NULL)
		return (0);
	ls = end = alloca (strlen (list) + 2);
	strcpy (end, list);
	if (timerange)
		cftime (clock, "%H%M", &now);
	else
		cftime (clock, "%Y%m%d", &now);
	present = atol (clock);
	while (*end)
	{	begin = end;
		if ((end = strchr (begin, '-')) == NULL)
			break;
		*end++ = '\0';
		first = atol (begin);
		last = atol (end);
		if (first <= present  &&  present <= last)
			return (0);
		if ((end = strchr (end, ',')) == NULL)
			break;
		end++;
	}
	if (timerange)
		fprintf (stderr, "Login is permitted between %s hours only.\n", list);
	else
		fprintf (stderr, "This login has been disabled.\n");
	return (1);
}

#endif

/****************************************************************************
*	set the users environment with values.									*
****************************************************************************/

void setenv (this, val)
char *this;
char *val;
{
	char	s[MAXPATHLEN+1];
	char	*ss;

	sprintf (s, "%s=%s", this, val);
	ss = malloc (strlen(s) + 1);
	strcpy (ss, s);
	putenv (ss);
}

/****************************************************************************
*	user-provided environment variables must be checked						*
****************************************************************************/

void check_putenv (val)
char *val;
{	static count = 0;
	char	*env;
	char	**forbid;
	
	/*	check forbidden environment variables */
	for (forbid = forbidden;  *forbid;  forbid++)
	{	if (strncmp (*forbid, val, strlen (*forbid)) == 0)
			return;
	}

	/*	 variables without an '=' sign assume the form of "Lnn=val" */
	if (strchr (val, '=') == NULL)
	{	env = malloc (strlen (val)+ 6);
		sprintf (env, "L%d=%s", count++, val);
		val = env;
	}
	putenv (val);
}

/****************************************************************************
*	get the encrypted password from /etc/shadow or /etc/d_passwd			*
*	allow "null" user for /etc/d_passwd general shell password.				*
*	Return encrypted password if found, else null.							*
****************************************************************************/

/*	could have used getspnam(user) except for /etc/d_passwd */

char *get_shadow (fname, user, allow)
char *fname;		/*	/etc/shadow or /etc/d_passwd						*/
char *user;			/*	user's login name									*/
int	 allow;			/*	0 =must have pwr entry, 1 =success even if no entry	*/
{	FILE	 *shadow;
	char	line[SHORTLINE+1];
	char	*pwd, *colon;
	static char pass[50];

	*pass = '\0';
	if (shadow = fopen (fname, "r"))
	{	while (fgets (line, SHORTLINE, shadow))
		{	line[SHORTLINE] = '\0';
			if (pwd = strchr (line, ':'))
			{	*pwd++ = '\0';
				if (colon = strchr (pwd, ':'))
				{	*colon++ = '\0';
					if (*line)
					{	if (strcmp (user, line) == 0)
						{	fclose (shadow);
							strcpy (pass, pwd);
							if (allow == 0  &&  colon)
							{	shadow_lstchg = atol (colon);
								colon = strchr (colon, ':') + 1;
								shadow_min = atol (colon);
								colon = strchr (colon, ':') + 1;
								shadow_max = atol (colon);
							}
							return (pass);
						}
					}
					else if (allow)
						strcpy (pass, pwd);
				}
			}
		}
	}
	fclose (shadow);
	if (allow  &&  *pass)
		return (pass);
	return (NULL);
}

/****************************************************************************
*	search for tty type in /etc/ttytype										*
****************************************************************************/

char *search_ttytype (tty)
char *tty;			/*	user's tty line										*/
{	FILE	*fp;
	char	line[SHORTLINE+1];
	static char *tname;
	char	*dev;
	
	if (fp = fopen (TTYTYPE, "r"))
	{	while (fgets (line, SHORTLINE, fp))
		{	line[SHORTLINE] = '\0';
			trim (line);
			dev = line;
			while (*dev)				/*	skip tty name 	*/
			{	if (*dev <= ' ')
					break;
				dev++;
			}
			while (*dev)				/*	locate device	*/
			{	if (*dev > ' ')
					break;
				*dev++ = '\0';
			}
			if (strcmp (dev, tty) == 0)
			{	fclose (fp);
				tname = malloc (strlen(line)+1);
				strcpy (tname, line);
				return (tname);
			}
		}
	}
	fclose (fp);
	return ("UNKNOWN");
}


/****************************************************************************
*	read /etc/default/login file											*
****************************************************************************/

int get_defaults (fname, tz, hz, console, ulim, passreq, altshell, 
				  path, supath, timeout, cmask, idleweeks)
char *fname;		/*	/etc/default/login									*/
char *tz;			/*	TIMEZONE=xxx	should be same as in /etc/TIMEZONE	*/
char *hz;			/*	HZ=nn												*/
char **console;		/*	CONSOLE=xxx     root must be on this device if given*/
long *ulim;			/*	ULIMIT=xxx		default ulimit for users			*/
int	*passreq;		/*	PASSREQ=YES/NO	is password required				*/
int *altshell;		/*	ALTSHELL=YES/NO	YES==set shell name in environment	*/
char **path;		/*	PATH=xxx		default user path					*/
char **supath;		/*	SUPATH=xxx		defaut root path					*/
long *timeout;		/*	TIMEOUT=nn		login abort timeout in seconds		*/
int  *cmask;		/*	UMASK=ooo		default umask for users				*/
int	 *idleweeks;	/*	IDLEWEEKS=nn	idle weeks							*/
{
	FILE	*fp;
	char	line[LONGLINE+1];
	char	*equal;
	
	if ((fp = fopen (fname, "r")) == NULL)
		return (2);
	while ((fgets (line, LONGLINE, fp)))
	{	if (*line == '#')				/*	must be first character */
			continue;
		line[LONGLINE] = '\0';
		trim (line);
		if (equal = strchr (line, '='))
		{	*equal++ = '\0';
			if (strcmp (line, "TIMEZONE") == 0)
				strcpy (tz, equal);
			else if (strcmp (line, "HZ") == 0)
				strcpy (hz, equal);
			else if (strcmp (line, "CONSOLE") == 0)
			{	if (*console == NULL)
				{	*console = malloc (strlen (equal) + 2);
					strcpy (*console, equal);
					if (*console[strlen(*console) - 1] != ':')
						strcat (*console, ":");
				}
			}
			else if (strcmp (line, "ULIMIT") == 0)
				*ulim = atol (equal);
			else if (strcmp (line, "PATH") == 0)
			{	if (*path == NULL)
				{	*path = malloc (strlen (equal) + 1);
					strcpy (*path, equal);
				}
			}
			else if (strcmp (line, "SUPATH") == 0)
			{	if (*supath == NULL)
				{	*supath = malloc (strlen (equal) + 1);
					strcpy (*supath, equal);
				}
			}
			else if (strcmp (line, "TIMEOUT") == 0)
				*timeout = atol (equal);
			else if (strcmp (line, "UMASK") == 0)
				*cmask = atoo (equal);
			else if (strcmp (line, "PASSREQ") == 0)
			{	if (toupper (*equal) == 'Y')
					*passreq = 1;
				else
					*passreq = 0;
			}
			else if (strcmp (line, "ALTSHELL") == 0)
			{	if (toupper (*equal) == 'Y')
					*altshell = 1;
				else
					*altshell = 0;
			}
			else if (strcmp (line, "IDLEWEEKS") == 0)
				*idleweeks = atoi (equal);
		}
	}
	fclose (fp);
	return (0);
}

/****************************************************************************
*	trim line at end														*
****************************************************************************/

void trim (s)
char *s;
{	char	*t;

	t = s + strlen (s);
	do
	{	if (*t > ' ')
			break;
		*t = '\0';
	} while (t-- != s);
}

/****************************************************************************
*	ascii to octal															*
****************************************************************************/

int atoo (s)
char *s;
{
	unsigned int v;
	
	v = 0;
	while (*s)
		v = v * 8 + (*s++ - '0');
	return (v);
}

/****************************************************************************
*	timeout error message if user takes too long to log in.					*
****************************************************************************/

void user_timeout()
{
	fprintf (stderr, "login timed out after %d seconds\n", timeout);
	exit(0);
}


/****************************************************************************
*	search for string t in string s.  return -1 if it does not exist		*
*	otherwise return an integer, the first byte of s that matches t.		*
****************************************************************************/

int tscan (s, t)
char 	s[], t[];
{
	int	i, j, k;
	for (i = 0;  s[i] != '\0';  i++)
	{	for (j = i, k=0;  t[k] != '\0'  &&  s[j] == t[k];  j++, k++)
			;
		if (t[k] == '\0')
			return (i);
	}
	return (-1);
}

/****************************************************************************
*	search for string t in string s.  return -1 if it does not exist		*
*	otherwise return an integer, the first byte of s that matches t.		*
*	Any '?' in s  is a wild card match for t								*
****************************************************************************/

int qtscan (s, t)
char 	s[], t[];
{
	int	i, j, k;
	for (i = 0;  s[i] != '\0';  i++)
	{	for (j = i, k=0;  t[k] != '\0'  &&  ((s[j] == t[k]) || (s[j] == '?')) ;  j++, k++)
			;
		if (t[k] == '\0')
			return (i);
	}
	return (-1);
}


